=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2008-11-03 21:52:46 +0000
+++ bzrlib/errors.py	2008-11-20 10:30:56 +0000
@@ -1547,11 +1547,15 @@
 
 class SmartMessageHandlerError(InternalBzrError):
 
-    _fmt = "The message handler raised an exception: %(exc_value)s."
+    _fmt = ("The message handler raised an exception:\n"
+            "%(traceback_text)s%(exc_value)s")
 
     def __init__(self, exc_info):
-        self.exc_type, self.exc_value, self.tb = exc_info
-        
+        import traceback
+        self.exc_type, self.exc_value, self.exc_tb = exc_info
+        self.exc_info = exc_info
+        self.traceback_text = ''.join(traceback.format_tb(self.exc_tb))
+
 
 # A set of semi-meaningful errors which can be thrown
 class TransportNotPossible(TransportError):

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2008-11-18 21:43:36 +0000
+++ bzrlib/remote.py	2008-11-20 10:30:56 +0000
@@ -18,6 +18,7 @@
 # across to run on the server.
 
 import bz2
+from cStringIO import StringIO
 
 from bzrlib import (
     branch,
@@ -41,6 +42,7 @@
 from bzrlib.smart import client, vfs
 from bzrlib.revision import ensure_null, NULL_REVISION
 from bzrlib.trace import mutter, note, warning
+from bzrlib.versionedfile import VersionedFiles
 
 
 class _RpcHelper(object):
@@ -345,6 +347,12 @@
         self.base = self.bzrdir.transport.base
         # Additional places to query for data.
         self._fallback_repositories = []
+        self.texts = RemoteVersionedFiles(self, 'texts')
+        self.inventories = RemoteVersionedFiles(self, 'inventories')
+        self.signatures = RemoteVersionedFiles(self, 'signatures')
+        self.revisions = RemoteVersionedFiles(self, 'revisions')
+        self._vf_objs = [
+            self.texts, self.inventories, self.signatures, self.revisions]
 
     def __str__(self):
         return "%s(%s)" % (self.__class__.__name__, self.base)
@@ -555,6 +563,8 @@
             self._lock_mode = 'r'
             self._lock_count = 1
             self._parents_map = {}
+            for vf in self._vf_objs:
+                vf._parents_map = {}
             if 'hpss' in debug.debug_flags:
                 self._requested_parents = set()
             if self._real_repository is not None:
@@ -597,6 +607,8 @@
             self._lock_mode = 'w'
             self._lock_count = 1
             self._parents_map = {}
+            for vf in self._vf_objs:
+                vf._parents_map = {}
             if 'hpss' in debug.debug_flags:
                 self._requested_parents = set()
         elif self._lock_mode == 'r':
@@ -664,6 +676,8 @@
         if self._lock_count > 0:
             return
         self._parents_map = None
+        for vf in self._vf_objs:
+            vf._parents_map = None
         if 'hpss' in debug.debug_flags:
             self._requested_parents = None
         old_mode = self._lock_mode
@@ -982,7 +996,6 @@
         included_keys = start_set.intersection(result_parents)
         start_set.difference_update(included_keys)
         recipe = (start_set, stop_keys, len(parents_map))
-        body = self._serialise_search_recipe(recipe)
         path = self.bzrdir._path_for_remote_call(self._client)
         for key in keys:
             if type(key) is not str:
@@ -992,7 +1005,7 @@
         args = (path,) + tuple(keys)
         try:
             response = self._call_with_body_bytes_expecting_body(
-                verb, args, body)
+                verb, args, _serialise_search_recipe(recipe))
         except errors.UnknownSmartMethod:
             # Server does not support this method, so get the whole graph.
             # Worse, we have to force a disconnection, because the server now
@@ -1113,16 +1126,6 @@
         # TODO: Suggestion from john: using external tar is much faster than
         # python's tarfile library, but it may not work on windows.
 
-    @property
-    def inventories(self):
-        """Decorate the real repository for now.
-
-        In the long term a full blown network facility is needed to
-        avoid creating a real repository object locally.
-        """
-        self._ensure_real()
-        return self._real_repository.inventories
-
     @needs_write_lock
     def pack(self):
         """Compress the data within the repository.
@@ -1132,47 +1135,15 @@
         self._ensure_real()
         return self._real_repository.pack()
 
-    @property
-    def revisions(self):
-        """Decorate the real repository for now.
-
-        In the short term this should become a real object to intercept graph
-        lookups.
-
-        In the long term a full blown network facility is needed.
-        """
-        self._ensure_real()
-        return self._real_repository.revisions
-
     def set_make_working_trees(self, new_value):
         self._ensure_real()
         self._real_repository.set_make_working_trees(new_value)
 
-    @property
-    def signatures(self):
-        """Decorate the real repository for now.
-
-        In the long term a full blown network facility is needed to avoid
-        creating a real repository object locally.
-        """
-        self._ensure_real()
-        return self._real_repository.signatures
-
     @needs_write_lock
     def sign_revision(self, revision_id, gpg_strategy):
         self._ensure_real()
         return self._real_repository.sign_revision(revision_id, gpg_strategy)
 
-    @property
-    def texts(self):
-        """Decorate the real repository for now.
-
-        In the long term a full blown network facility is needed to avoid
-        creating a real repository object locally.
-        """
-        self._ensure_real()
-        return self._real_repository.texts
-
     @needs_read_lock
     def get_revisions(self, revision_ids):
         self._ensure_real()
@@ -1228,16 +1199,207 @@
                          self._fallback_repositories)
         return graph._StackedParentsProvider(providers)
 
-    def _serialise_search_recipe(self, recipe):
-        """Serialise a graph search recipe.
-
-        :param recipe: A search recipe (start, stop, count).
-        :return: Serialised bytes.
-        """
-        start_keys = ' '.join(recipe[0])
-        stop_keys = ' '.join(recipe[1])
-        count = str(recipe[2])
-        return '\n'.join((start_keys, stop_keys, count))
+def _serialise_search_recipe(recipe):
+    """Serialise a graph search recipe.
+
+    :param recipe: A search recipe (start, stop, count).
+    :return: Serialised bytes.
+    """
+    start_keys = ' '.join(recipe[0])
+    stop_keys = ' '.join(recipe[1])
+    count = str(recipe[2])
+    return '\n'.join((start_keys, stop_keys, count))
+
+def _serialise_search_tuple_key_recipe((start, stop, count)):
+    """Serialise a graph search recipe.
+
+    :param recipe: A search recipe (start, stop, count).
+    :return: Serialised bytes.
+    """
+    # Format:
+    #  recipe = keys NEWLINE keys NEWLINE count
+    #  count = INTEGER
+    #  keys = key 0x00 key 0x00 key ...
+    #  key = key_elem SPACE key_elem SPACE key_elem ...
+    buf = StringIO()
+    start_keys = (' '.join(key) for key in start)
+    buf.write('\0'.join(start_keys))
+    buf.write('\n')
+    stop_keys = (' '.join(key) for key in stop)
+    buf.write('\0'.join(stop_keys))
+    buf.write('\n')
+    buf.write(str(count))
+    return buf.getvalue()
+
+
+def _deserialise_search_result(bytes):
+    graph_dict = {}
+    for line in bytes.split('\n'):
+        keys_bytes = line.split('\0')
+        keys = [tuple(key_bytes.split(' ')) for key_bytes in keys_bytes]
+        parents = tuple(keys[1:])
+        if parents == (('',),):
+            parents = ()
+        graph_dict[keys[0]] = parents
+    return graph_dict
+
+
+class RemoteVersionedFiles(VersionedFiles):
+
+    def __init__(self, remote_repo, vf_name):
+        self.remote_repo = remote_repo
+        self.vf_name = vf_name
+        self._parents_map = None
+        if 'hpss' in debug.debug_flags:
+            self._requested_parents = None
+
+    def _get_real_vf(self):
+        self.remote_repo._ensure_real()
+        return getattr(self.remote_repo._real_repository, self.vf_name)
+
+    def add_lines(self, version_id, parents, lines, parent_texts=None,
+            left_matching_blocks=None, nostore_sha=None, random_id=False,
+            check_content=True):
+        real_vf = self._get_real_vf()
+        return real_vf.add_lines(version_id, parents, lines,
+            parent_texts=parent_texts,
+            left_matching_blocks=left_matching_blocks,
+            nostore_sha=nostore_sha, random_id=random_id,
+            check_content=check_content)
+
+    def add_mpdiffs(self, records):
+        real_vf = self._get_real_vf()
+        return real_vf.add_mpdiffs(records)
+
+    def annotate(self, key):
+        real_vf = self._get_real_vf()
+        return real_vf.annotate(key)
+
+    def check(self, progress_bar=None):
+        real_vf = self._get_real_vf()
+        return real_vf.check(progress_bar=progress_bar)
+
+    def get_parent_map(self, keys):
+        # XXX: lots of duplication with RemoteRepository.get_parent_map.
+        client = self.remote_repo._client
+        medium = client._medium
+        if medium._is_remote_before((1, 8)):
+            real_vf = self._get_real_vf()
+            return real_vf.get_parent_map(keys)
+        # Hack to build up the caching logic.
+        ancestry = self._parents_map
+        if ancestry is None:
+            # Repository is not locked, so there's no cache.
+            missing_revisions = set(keys)
+            ancestry = {}
+        else:
+            missing_revisions = set(key for key in keys if key not in ancestry)
+        if missing_revisions:
+            try:
+                parent_map = self._get_parent_map(missing_revisions)
+            except errors.UnknownSmartMethod:
+                medium._remember_remote_is_before((1, 8))
+                real_vf = self._get_real_vf()
+                return real_vf.get_parent_map(keys)
+            if 'hpss' in debug.debug_flags:
+                mutter('retransmitted revisions: %d of %d',
+                        len(set(ancestry).intersection(parent_map)),
+                        len(parent_map))
+            ancestry.update(parent_map)
+        present_keys = [k for k in keys if k in ancestry]
+        if 'hpss' in debug.debug_flags:
+            if self._requested_parents is not None and len(ancestry) != 0:
+                self._requested_parents.update(present_keys)
+                mutter('Current RemoteRepository graph hit rate: %d%%',
+                    100.0 * len(self._requested_parents) / len(ancestry))
+        return dict((k, ancestry[k]) for k in present_keys)
+        
+    def _get_parent_map(self, keys):
+        # XXX: lots of duplication with RemoteRepository._get_parent_map.
+        if None in keys:
+            raise ValueError('get_parent_map(None) is not valid')
+        if NULL_REVISION in keys:
+            keys.discard(NULL_REVISION)
+            found_parents = {NULL_REVISION:()}
+            if not keys:
+                return found_parents
+        else:
+            found_parents = {}
+        # TODO(Needs analysis): We could assume that the keys being requested
+        # from get_parent_map are in a breadth first search, so typically they
+        # will all be depth N from some common parent, and we don't have to
+        # have the server iterate from the root parent, but rather from the
+        # keys we're searching; and just tell the server the keyspace we
+        # already have; but this may be more traffic again.
+
+        # Transform self._parents_map into a search request recipe.
+        # TODO: Manage this incrementally to avoid covering the same path
+        # repeatedly. (The server will have to on each request, but the less
+        # work done the better).
+        parents_map = self._parents_map
+        if parents_map is None:
+            # Repository is not locked, so there's no cache.
+            parents_map = {}
+        start_set = set(parents_map)
+        result_parents = set()
+        for parents in parents_map.itervalues():
+            result_parents.update(parents)
+        stop_keys = result_parents.difference(start_set)
+        included_keys = start_set.intersection(result_parents)
+        start_set.difference_update(included_keys)
+        recipe = (start_set, stop_keys, len(parents_map))
+        client = self.remote_repo._client
+        path = self.remote_repo.bzrdir._path_for_remote_call(client)
+        for key in keys:
+            if type(key) is not tuple:
+                raise ValueError(
+                    "key %r not a tuple of strs" % (key,))
+            for key_elem in key:
+                if type(key_elem) is not str:
+                    raise ValueError(
+                        "key %r not a tuple of strs" % (key,))
+        verb = 'VersionedFiles.get_parent_map'
+        args = (path, self.vf_name) + tuple(keys)
+        response = client.call_with_body_bytes_expecting_body(
+            verb, args, _serialise_search_tuple_key_recipe(recipe))
+        response_tuple, response_handler = response
+        if response_tuple[0] not in ['ok']:
+            response_handler.cancel_read_body()
+            raise errors.UnexpectedSmartServerResponse(response_tuple)
+        if response_tuple[0] == 'ok':
+            coded = bz2.decompress(response_handler.read_body_bytes())
+            revision_graph = _deserialise_search_result(coded)
+            # The server doesn't return explicit results for keys it doesn't
+            # have, so add empty entries for the missing keys to the result
+            # dict.
+            missing = keys.difference(revision_graph)
+            for missing_key in missing:
+                revision_graph[missing_key] = ()
+            return revision_graph
+
+    def get_record_stream(self, keys, ordering, include_delta_closure):
+        real_vf = self._get_real_vf()
+        return real_vf.get_record_stream(keys, ordering, include_delta_closure)
+
+    def get_sha1s(self, keys):
+        real_vf = self._get_real_vf()
+        return real_vf.get_sha1s(keys)
+
+    def insert_record_stream(self, stream):
+        real_vf = self._get_real_vf()
+        return real_vf.insert_record_stream(stream)
+
+    def iter_lines_added_or_present_in_keys(self, keys, pb=None):
+        real_vf = self._get_real_vf()
+        return real_vf.iter_lines_added_or_present_in_keys(keys, pb=pb)
+
+    def keys(self):
+        real_vf = self._get_real_vf()
+        return real_vf.keys()
+
+    def make_mpdiffs(self, keys):
+        real_vf = self._get_real_vf()
+        return real_vf.make_mpdiffs(keys)
 
     def autopack(self):
         path = self.bzrdir._path_for_remote_call(self._client)

=== modified file 'bzrlib/smart/client.py'
--- bzrlib/smart/client.py	2008-09-24 13:46:44 +0000
+++ bzrlib/smart/client.py	2008-11-20 10:30:56 +0000
@@ -151,9 +151,9 @@
         """Call a method on the remote server with body bytes."""
         if type(method) is not str:
             raise TypeError('method must be a byte string, not %r' % (method,))
-        for arg in args:
-            if type(arg) is not str:
-                raise TypeError('args must be byte strings, not %r' % (args,))
+#        for arg in args:
+#            if type(arg) is not str:
+#                raise TypeError('args must be byte strings, not %r' % (args,))
         if type(body) is not str:
             raise TypeError('body must be byte string, not %r' % (body,))
         response, response_handler = self._call_and_read_response(

=== modified file 'bzrlib/smart/message.py'
--- bzrlib/smart/message.py	2008-10-01 06:54:56 +0000
+++ bzrlib/smart/message.py	2008-11-20 10:30:56 +0000
@@ -281,6 +281,7 @@
             body_bytes = ''.join(self._bytes_parts)
             if 'hpss' in debug.debug_flags:
                 mutter('              %d body bytes read', len(body_bytes))
+                mutter('              body %r...', repr(body_bytes[:40]))
             self._body = StringIO(body_bytes)
             self._bytes_parts = None
         return self._body.read(count)

=== modified file 'bzrlib/smart/protocol.py'
--- bzrlib/smart/protocol.py	2008-09-17 07:18:20 +0000
+++ bzrlib/smart/protocol.py	2008-09-24 01:52:45 +0000
@@ -758,6 +758,8 @@
         if 'hpss' in debug.debug_flags:
             mutter('              %d body bytes read',
                    len(self._body_buffer.getvalue()))
+            mutter('              body %r...',
+                   repr(self._body_buffer.getvalue()[:40]))
         return self._body_buffer.read(count)
 
     def _recv_tuple(self):
@@ -1182,7 +1184,7 @@
         After calling this, call read_response_tuple to find the result out.
         """
         if 'hpss' in debug.debug_flags:
-            mutter('hpss call w/body: %s (%r...)', repr(args)[1:-1], body[:20])
+            mutter('hpss call w/body: %s (%r...)', repr(args)[1:-1], body[:200])
             path = getattr(self._medium_request._medium, '_path', None)
             if path is not None:
                 mutter('                  (to %s)', path)

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2008-10-30 03:47:59 +0000
+++ bzrlib/smart/request.py	2008-11-20 10:30:56 +0000
@@ -473,3 +473,6 @@
     'Transport.is_readonly', 'bzrlib.smart.request', 'SmartServerIsReadonly')
 request_handlers.register_lazy(
     'BzrDir.open', 'bzrlib.smart.bzrdir', 'SmartServerRequestOpenBzrDir')
+request_handlers.register_lazy('VersionedFiles.get_parent_map',
+                               'bzrlib.smart.versionedfiles',
+                               'SmartServerVersionedFilesGetParentMap')

=== added file 'bzrlib/smart/versionedfiles.py'
--- bzrlib/smart/versionedfiles.py	1970-01-01 00:00:00 +0000
+++ bzrlib/smart/versionedfiles.py	2008-09-22 09:33:04 +0000
@@ -0,0 +1,196 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Server-side versioned files related request implmentations."""
+
+import bz2
+from cStringIO import StringIO
+
+from bzrlib import (
+    errors,
+    graph,
+    )
+from bzrlib.bzrdir import BzrDir
+from bzrlib.smart.request import (
+    FailedSmartServerResponse,
+    SmartServerRequest,
+    SuccessfulSmartServerResponse,
+    )
+from bzrlib.smart.repository import SmartServerRepositoryRequest
+from bzrlib.repository import _strip_NULL_ghosts
+from bzrlib import revision as _mod_revision
+
+
+class SmartServerVersionedFilesRequest(SmartServerRepositoryRequest):
+    """Common base class for Repository requests."""
+
+    def do(self, path, *args):
+        """Execute a repository request.
+        
+        All Repository requests take a path to the repository as their first
+        argument.  The repository must be at the exact path given by the
+        client - no searching is done.
+
+        The actual logic is delegated to self.do_repository_request.
+
+        :param client_path: The path for the repository as received from the
+            client.
+        :return: A SmartServerResponse from self.do_repository_request().
+        """
+        transport = self.transport_from_client_path(path)
+        bzrdir = BzrDir.open_from_transport(transport)
+        # Save the repository for use with do_body.
+        self._repository = bzrdir.open_repository()
+        return self.do_repository_request(self._repository, *args)
+
+    def do_repository_request(self, repository, vf_name, *args):
+        if vf_name not in ['texts', 'inventories', 'signatures', 'revisions']:
+            return FailedSmartServerResponse(('NoSuchVersionedFile', vf_name))
+        versioned_files = getattr(repository, vf_name)
+        return self.do_versioned_files_request(versioned_files, *args)
+
+    def do_versioned_files_request(self, versioned_files, *args):
+        """Override to provide an implementation for a verb."""
+
+
+class SmartServerVersionedFilesGetParentMap(SmartServerVersionedFilesRequest):
+    """Bzr 1.2+ - get parent data for revisions during a graph search."""
+    
+    def do_versioned_files_request(self, versioned_files, *keys):
+        """Get parent details for some revisions.
+        
+        All the parents for revision_ids are returned. Additionally up to 64KB
+        of additional parent data found by performing a breadth first search
+        from revision_ids is returned. The verb takes a body containing the
+        current search state, see do_body for details.
+
+        :param repository: The repository to query in.
+        :param revision_ids: The utf8 encoded revision_id to answer for.
+        """
+        self._keys = tuple(tuple(key) for key in keys)
+        self._versioned_files = versioned_files
+        return None # Signal that we want a body.
+
+    def do_body(self, body_bytes):
+        """Process the current search state and perform the parent lookup.
+
+        :return: A smart server response where the body contains an utf8
+            encoded flattened list of the parents of the keys which has been
+            bz2 compressed.
+        """
+        repository = self._repository
+        repository.lock_read()
+        try:
+            vf_graph = graph.Graph(self._versioned_files)
+            return self._do_repository_request(body_bytes, vf_graph)
+        finally:
+            repository.unlock()
+
+    def _deserialise_search_tuple_key_recipe(self, bytes):
+        start_keys_bytes, stop_keys_bytes, count_bytes = bytes.split('\n')
+        start_keys = [tuple(k.split(' ')) for k in start_keys_bytes.split('\0')]
+        stop_keys = [tuple(k.split(' ')) for k in stop_keys_bytes.split('\0')]
+        count = int(count_bytes)
+        return tuple(start_keys), set(stop_keys), count
+
+    def recreate_vf_search(self, vf_graph, recipe_bytes):
+        recipe = self._deserialise_search_tuple_key_recipe(recipe_bytes)
+        start_keys, exclude_keys, key_count = recipe
+        # lock_read
+        try:
+            search = vf_graph._make_breadth_first_searcher(start_keys)
+            while True:
+                try:
+                    next_revs = search.next()
+                except StopIteration:
+                    break
+                search.stop_searching_any(exclude_keys.intersection(next_revs))
+            search_result = search.get_result()
+            if search_result.get_recipe()[2] != key_count:
+                # we got back a different amount of data than expected, this
+                # gets reported as NoSuchRevision, because less revisions
+                # indicates missing revisions, and more should never happen as
+                # the excludes list considers ghosts and ensures that ghost
+                # filling races are not a problem.
+                return (None, FailedSmartServerResponse(('NoSuchRevision',)))
+            return (search, None)
+        finally:
+            pass
+            # unlock
+            
+    def _do_repository_request(self, body_bytes, vf_graph):
+        repository = self._repository
+        keys = set(self._keys)
+        search, error = self.recreate_vf_search(vf_graph, body_bytes)
+        if error is not None:
+            return error
+        # TODO might be nice to start up the search again; but thats not
+        # written or tested yet.
+        client_seen_keys = set(search.get_result().get_keys())
+        # Always include the requested ids.
+        client_seen_keys.difference_update(keys)
+        lines = []
+        result = {}
+        queried_keys = set()
+        size_so_far = 0
+        next_keys = keys
+        first_loop_done = False
+        while next_keys:
+            queried_keys.update(next_keys)
+            parent_map = vf_graph.get_parent_map(next_keys)
+            next_keys = set()
+            for key, parents in parent_map.iteritems():
+                # prepare the next query
+                next_keys.update(parents)
+                if key not in client_seen_keys:
+                    # Client does not have this revision, give it to it.
+                    # add parents to the result
+                    result[key] = parents
+                    # Approximate the serialized cost of this key.
+                    # XXX: this approximation is out of date
+                    size_so_far += 2 + len(key) + sum(map(len, parents))
+            # get all the directly asked for parents, and then flesh out to
+            # 64K (compressed) or so. We do one level of depth at a time to
+            # stay in sync with the client. The 250000 magic number is
+            # estimated compression ratio taken from bzr.dev itself.
+            if first_loop_done and size_so_far > 250000:
+                next_keys = set()
+                break
+            # don't query things we've already queried
+            next_keys.difference_update(queried_keys)
+            first_loop_done = True
+
+        # sorting trivially puts lexographically similar revision ids together.
+        # Compression FTW.
+        result = sorted(result.items())
+
+        return SuccessfulSmartServerResponse(
+            ('ok', ), bz2.compress(_serialise_search_result(result)))
+
+
+def _serialise_search_result(result_items):
+    buf = StringIO()
+    first = True
+    for key, parents in result_items:
+        if first:
+            first = False
+        else:
+            buf.write('\n')
+        buf.write(' '.join(key))
+        buf.write('\0')
+        parents_iter = (' '.join(parent) for parent in parents)
+        buf.write('\0'.join(parents_iter))
+    return buf.getvalue()

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2008-11-18 21:43:36 +0000
+++ bzrlib/tests/test_remote.py	2008-11-20 10:30:56 +0000
@@ -1168,7 +1168,7 @@
     def test_get_parent_map_reconnects_if_unknown_method(self):
         transport_path = 'quack'
         repo, client = self.setup_fake_client_and_repository(transport_path)
-        client.add_unknown_method_response('Repository,get_parent_map')
+        client.add_unknown_method_response('Repository.get_parent_map')
         client.add_success_response_with_body('', 'ok')
         self.assertFalse(client._medium._is_remote_before((1, 2)))
         rev_id = 'revision-id'
@@ -1509,6 +1509,72 @@
             client._calls)
 
 
+class TestVersionedFilesGetParentMap(tests.TestCaseWithTransport):
+
+    def test_backwards_compat(self):
+        """If the server does not recognise the VersionedFiles.get_parent_map
+        then the client will fallback to calling get_parent_map on the 'real'
+        versioned files object.
+        """
+        self.transport_server = server.SmartTCPServer_for_testing
+        self.make_repository('branch')
+        repo = repository.Repository.open(self.get_url('branch'))
+        repo._ensure_real()
+        repo.lock_read()
+        # Sabotage the real smart medium so that the test will fail if
+        # something tries to use it after this point.
+        repo._client._medium._socket = None
+        # Replace the _real_repository with a test double so that we can
+        # observe the fallback to using the 'real' object.
+        class FakeRealVf(object):
+            def get_parent_map(self, keys):
+                client._calls.append(('real vf get_parent_map', keys))
+                return {}
+        class FakeRealRepo(object):
+            def __init__(self):
+                self.revisions = FakeRealVf()
+            def unlock(self):
+                pass
+        repo._real_repository = FakeRealRepo()
+        # Inject a test double client.
+        client = FakeClient(repo.bzrdir.root_transport.base)
+        repo._client = client
+        # Configure the fake client.
+        client.add_unknown_method_response('VersionedFiles.get_parent_map')
+        client.add_success_response_with_body('', 'ok')
+        # Call get_parent_map.
+        rev_id = 'revision-id'
+        repo.revisions.get_parent_map([('rev-id',)])
+        repo.unlock()
+        self.assertEqual(
+            [('call_with_body_bytes_expecting_body',
+              'VersionedFiles.get_parent_map',
+              ('extra/branch/', 'revisions', ('rev-id',)),
+              '\n\n0'),
+             ('real vf get_parent_map', [('rev-id',)]),
+             ],
+            client._calls)
+        # The medium is now marked as being connected to an older server
+        self.assertTrue(client._medium._is_remote_before((1, 8)))
+
+
+class TestVersionedFilesGetParentMapSerialisation(tests.TestCase):
+
+    def test_serialise_search_recipe_empty(self):
+        recipe = ([], [], 0)
+        bytes = remote._serialise_search_tuple_key_recipe(recipe)
+        self.assertEqual('\n\n0', bytes)
+
+    def test_serialise_search_recipe(self):
+        start = [('start', 'key', 'one'), ('start', 'key', 'two')]
+        stop = [('stop', 'key', 'one'), ('stop', 'key', 'two')]
+        recipe = (start, stop, 123)
+        bytes = remote._serialise_search_tuple_key_recipe(recipe)
+        self.assertEqual(
+            'start key one\x00start key two\nstop key one\x00stop key two\n123',
+            bytes)
+
+
 class TestErrorTranslationBase(tests.TestCaseWithMemoryTransport):
     """Base class for unit tests for bzrlib.remote._translate_error."""
 

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2008-10-31 01:57:16 +0000
+++ bzrlib/tests/test_smart.py	2008-11-20 10:30:56 +0000
@@ -40,6 +40,7 @@
 import bzrlib.smart.branch
 import bzrlib.smart.bzrdir
 import bzrlib.smart.repository
+import bzrlib.smart.versionedfiles
 from bzrlib.smart.request import (
     FailedSmartServerResponse,
     SmartServerRequest,
@@ -1084,6 +1085,23 @@
         self.assertEqual(SmartServerResponse(('ok',)), response)
         
 
+class TestSmartServerVersionedFilesGetParentMap(tests.TestCaseWithMemoryTransport):
+
+    def test_trivial_bzipped(self):
+        # This tests that the wire encoding is actually bzipped
+        backing = self.get_transport()
+        request = smart.versionedfiles.SmartServerVersionedFilesGetParentMap(
+            backing)
+        tree = self.make_branch_and_memory_tree('.')
+
+        self.assertEqual(
+            None, request.execute('', 'revisions', ('missing-id',)))
+        # Note that it returns a body (of '' bzipped).
+        self.assertEqual(
+            SuccessfulSmartServerResponse(('ok', ), bz2.compress('')),
+            request.do_body('\n\n0'))
+
+
 class TestHandlers(tests.TestCase):
     """Tests for the request.request_handlers object."""
 

